from dotgraph import Node, dot_output, invoke_dot, invoke_dot_aa, NoDot, NoRsvg
from dotgraph import RSVG_OUTPUT_TYPES, DOT_OUTPUT_TYPES, Edge
from bzrlib.branch import Branch
from bzrlib.errors import BzrCommandError, NoCommonRoot, NoSuchRevision
from bzrlib.fetch import greedy_fetch
from bzrlib.graph import node_distances, select_farthest
from bzrlib.revision import combined_graph, revision_graph
from bzrlib.revision import MultipleRevisionSources
from datetime import datetime
import bzrlib.errors
import re
import os.path

mail_map = {'aaron.bentley@utoronto.ca'     : 'Aaron Bentley',
            'abentley@panoramicfeedback.com': 'Aaron Bentley',
            'abentley@lappy'                : 'Aaron Bentley',
            'john@arbash-meinel.com'        : 'John Arbash Meinel',
            'mbp@sourcefrog.net'            : 'Martin Pool',
            'robertc@robertcollins.net'     : 'Robert Collins',
            }

committer_alias = {'abentley': 'Aaron Bentley'}
def short_committer(committer):
    new_committer = re.sub('<.*>', '', committer).strip(' ')
    if len(new_committer) < 2:
        return committer
    return new_committer

def can_skip(rev_id, descendants, ancestors):
    if rev_id not in descendants:
        return False
    elif rev_id not in ancestors:
        return False
    elif len(ancestors[rev_id]) != 1:
        return False
    elif len(descendants[list(ancestors[rev_id])[0]]) != 1:
        return False
    elif len(descendants[rev_id]) != 1:
        return False
    else:
        return True

def compact_ancestors(descendants, ancestors, exceptions=()):
    new_ancestors={}
    skip = set()
    for me, my_parents in ancestors.iteritems():
        if me in skip:
            continue
        new_ancestors[me] = {} 
        for parent in my_parents:
            new_parent = parent 
            distance = 0
            while can_skip(new_parent, descendants, ancestors):
                if new_parent in exceptions:
                    break
                skip.add(new_parent)
                if new_parent in new_ancestors:
                    del new_ancestors[new_parent]
                new_parent = list(ancestors[new_parent])[0]
                distance += 1
            new_ancestors[me][new_parent] = distance
    return new_ancestors    

def get_committer(rev_id, source):
    try:
        committer = short_committer(source.get_revision(rev_id).committer)
    except NoSuchRevision:
        try:
            committer = '-'.join(rev_id.split('-')[:-2]).strip(' ')
            if committer == '':
                return None
        except ValueError:
            return None
    if '@' in committer:
        try:
            committer = mail_map[committer]
        except KeyError:
            pass
    try:
        committer = committer_alias[committer]
    except KeyError:
        pass
    return committer

class Grapher(object):
    def __init__(self, branch, other_branch=None):
        object.__init__(self)
        self.branch = branch
        self.other_branch = other_branch
        revision_a = self.branch.last_patch()
        if other_branch is not None:
            greedy_fetch(branch, other_branch)
            revision_b = self.other_branch.last_patch()
            try:
                self.root, self.ancestors, self.descendants, self.common = \
                    combined_graph(revision_a, revision_b, self.branch)
            except bzrlib.errors.NoCommonRoot:
                raise bzrlib.errors.NoCommonAncestor(revision_a, revision_b)
        else:
            self.root, self.ancestors, self.descendants = \
                revision_graph(revision_a, branch)
            self.common = []

        self.n_history = branch.revision_history()
        self.distances = node_distances(self.descendants, self.ancestors, 
                                        self.root)
        if other_branch is not None:
            self.base = select_farthest(self.distances, self.common)
            self.m_history = other_branch.revision_history() 
        else:
            self.base = None
            self.m_history = []

    def get_timestamp(self, revision_id):
        try:
            return float(self.branch.get_revision(revision_id).timestamp)
        except NoSuchRevision:
            return None

    def dot_node(self, node, num):
        try:
            n_rev = self.n_history.index(node) + 1
        except ValueError:
            n_rev = None
        try:
            m_rev = self.m_history.index(node) + 1
        except ValueError:
            m_rev = None
        if (n_rev, m_rev) == (None, None):
            name = node[-5:]
            cluster = None
        elif n_rev == m_rev:
            name = "rR%d" % n_rev
        else:
            namelist = []
            for prefix, revno in (('r', n_rev), ('R', m_rev)):
                if revno is not None:
                    namelist.append("%s%d" % (prefix, revno))
            name = ' '.join(namelist)
        if None not in (n_rev, m_rev):
            cluster = "common_history"
            color = "#ff9900"
        elif (None, None) == (n_rev, m_rev):
            cluster = None
            if node in self.common:
                color = "#6699ff"
            else:
                color = "white"
        elif n_rev is not None:
            cluster = "my_history"
            color = "#ffff00"
        else:
            assert m_rev is not None
            cluster = "other_history"
            color = "#ff0000"
        if node == self.base:
            color = "#33ff99"

        label = [name]
        committer = get_committer(node, self.branch)
        if committer is not None:
            label.append(committer)

        timestamp = self.get_timestamp(node)
        if timestamp is not None:
            date = datetime.fromtimestamp(timestamp)
            label.append("%d/%d/%d" % (date.year, date.month, date.day))

        if node in self.distances:
            rank = self.distances[node]
            label.append('d%d' % self.distances[node])
        else:
            rank = None

        d_node = Node("n%d" % num, color=color, label="\\n".join(label), 
                    rev_id=node, cluster=cluster)
        d_node.rank = rank

        if node not in self.ancestors:
            d_node.node_style.append('dotted')
        d_node.href = '#'

        return d_node
        
    def get_relations(self, collapse=False):
        dot_nodes = {}
        node_relations = []
        num = 0
        if collapse:
            visible_ancestors = compact_ancestors(self.descendants, 
                                                  self.ancestors, (self.base,))
        else:
            visible_ancestors = self.ancestors
        for node, parents in visible_ancestors.iteritems():
            if node not in dot_nodes:
                dot_nodes[node] = self.dot_node(node, num)
                num += 1
            if visible_ancestors is self.ancestors:
                parent_iter = ((f, 0) for f in parents)
            else:
                parent_iter = (f for f in parents.iteritems())
            for parent, skipped in parent_iter:
                if parent not in dot_nodes:
                    dot_nodes[parent] = self.dot_node(parent, num)
                    num += 1
                edge = Edge(dot_nodes[parent], dot_nodes[node])
                if skipped != 0:
                    edge.label = "%d" % skipped
                node_relations.append(edge)
        return node_relations


def write_ancestry_file(branch, filename, collapse=True, antialias=True,
                        merge_branch=None, ranking="forced"):
    b = Branch.open_containing(branch)
    if merge_branch is not None:
        m = Branch.open_containing(merge_branch)
    else:
        m = None
    grapher = Grapher(b, m)
    relations = grapher.get_relations(collapse)

    ext = filename.split('.')[-1]
    output = dot_output(relations, ranking)
    done = False
    if ext not in RSVG_OUTPUT_TYPES:
        antialias = False
    if antialias: 
        output = list(output)
        try:
            invoke_dot_aa(output, filename, ext)
            done = True
        except NoDot, e:
            raise BzrCommandError("Can't find 'dot'.  Please ensure Graphviz"\
                " is installed correctly.")
        except NoRsvg, e:
            print "Not antialiasing because rsvg (from librsvg-bin) is not"\
                " installed."
            antialias = False
    if ext in DOT_OUTPUT_TYPES and not antialias and not done:
        try:
            invoke_dot(output, filename, ext)
            done = True
        except NoDot, e:
            raise BzrCommandError("Can't find 'dot'.  Please ensure Graphviz"\
                " is installed correctly, or use --noantialias")
    elif ext=='dot' and not done:
        my_file = file(filename, 'wb')
        for fragment in output:
            my_file.write(fragment)
    elif not done:
        print "Unknown file extension: %s" % ext

